#include <setjmp.h>
#include "data.h"

void readFEN(char gen[]);
void showFullBoard();
void generate_moves();
short int make_move(move moveDetails);
short int take_back();
void showGenMoves();
void showHistory();
void terminate();
void sort_pv();
void sort(int fromMove);
int ProbeHash(int depth, int alpha, int beta);
void RecordHash(int depth, int value, int flags);

jmp_buf env;
short int done;

void think()
{
	int i,x;

	done = 0;
	setjmp(env);

	if (done)
	{
		while (ply) take_back();
		return;
	}

	start_time = getTime();
	stop_time = start_time + max_time;

	ply=0;
	nodes=0;

	memset(pv, 0, sizeof(pv));

	for (i = 1; i <= max_depth; ++i)
	{
		follow_pv = 1;
		x = NegaMax(i,-10000, 10000);
		printf("%3d %5d %5d %9d ", i, x, getTime()-start_time, nodes);
		showPV();
		if (x > 10000 || x < -10000)
			break;
	}
	return;
}

Perft(int depth)
{
	if (!depth) return;
	generate_moves();
	if (first_move[ply]==first_move[ply + 1]) return;
	int i;
	short int moved=0;
	for (i = first_move[ply]; i < first_move[ply + 1]; ++i)
	{
		if (!make_move(move_list[i])) continue;
		moved=1;
		if (depth==1)
{
		perft++;
}
		Perft(depth-1);
		take_back();
	}
}

int NegaMax(int depth, int alpha, int beta)
{
	int hashf = hashfALPHA;

	int i, x;
	short int moved;

	if ((x = ProbeHash(depth,alpha,beta)) != valUNKNOWN)
		return x;

	if (!depth)
	{
		x = NegaMaxQuiescent(alpha,beta);//evaluate();
		RecordHash(depth,x,hashfEXACT);
		return x;
	}

	nodes++;

	if (nodes&4096) if (mode==XBOARD) terminate();

	pv_length[ply] = ply;

	if (ply >= MAX_PLY - 1)
		return evaluate();
	if (hply >= HIST_STACK - 1)
		return evaluate();

	generate_moves();
	if (follow_pv)  /* are we following the PV? */
		sort_pv();
	moved = 0;


	if (first_move[ply]==first_move[ply + 1])
	{
		side=!side;
		if (attacked(kings[!side]))
		{
			side=!side;
			return -100000;	//checkmate
		}
		else
		{
			side=!side;
			return 0;	//stalemate
		}
	}

	/* loop through the moves */
	for (i = first_move[ply]; i < first_move[ply + 1]; ++i) {
		sort(i);
		if (!make_move(move_list[i])) continue;
		moved = 1;
		x = -NegaMax(depth-1,-beta, -alpha);
		take_back();
		if (x > alpha) {

			if (x >= beta)
			{
				RecordHash(depth, beta, hashfBETA);
				return beta;
			}
			hashf = hashfEXACT;
			alpha = x;

		pv[ply][ply]=move_list[i];

		int j;
		for (j = ply + 1; j < pv_length[ply + 1]; ++j)
			pv[ply][j]=pv[ply+1][j];

		pv_length[ply] = pv_length[ply + 1];

		}
	}
	if (!moved)
	{
		side=!side;
		if (attacked(kings[!side]))
		{
			side=!side;
			return -100000;	//checkmate
		}
		else
		{
			side=!side;
			return 0;	//stalemate
		}
	}
	RecordHash(depth, alpha, hashf);
	return alpha;
}

int NegaMaxQuiescent(int alpha, int beta)
{
	int i, x;
	short int moved;

	nodes++;
	if (nodes&4096) if (mode==XBOARD) terminate();

	pv_length[ply] = ply;

	if (ply >= MAX_PLY - 1)
		return evaluate();
	if (hply >= HIST_STACK - 1)
		return evaluate();

	x = evaluate();
	if (x >= beta)
		return beta;
	if (x > alpha)
		alpha = x;


	generate_captures();
	if (follow_pv)
		sort_pv();

	for (i = first_move[ply]; i < first_move[ply + 1]; ++i) {
		sort(i);
		if (!make_move(move_list[i])) continue;
		moved = 1;
		x = -NegaMaxQuiescent(-beta, -alpha);
		take_back();
		if (x > alpha) {

			if (x >= beta)
				return beta;
			alpha = x;

		//update PV

		pv[ply][ply]=move_list[i];
		int j;
		for (j = ply + 1; j < pv_length[ply + 1]; ++j)
			pv[ply][j]=pv[ply+1][j];
		pv_length[ply] = pv_length[ply + 1];

		}
	}

	return alpha;
}






int NegaScout(int depth,int alpha,int beta)
{
	if (depth <= 0) return evaluate();
	pv_length[ply] = ply;

	if (ply >= MAX_PLY - 1)
		return evaluate();

	int b = beta;
	int i,k;
	generate_moves();
	if (follow_pv)  /* are we following the PV? */
		sort_pv();
	for (i = k = first_move[ply]; i < first_move[ply + 1]; i++)
	{
		sort(i);
		if (!make_move(move_list[i])) continue;
		showHistory();
		int temp=-NegaScout(depth-1,-b,-alpha);
		if((alpha<temp)&&(temp<beta)&&(i>k)) //and not the first child
		{
			alpha=-NegaScout(depth-1,-beta,-temp); /* re-search */
		}
		take_back();
		alpha=alpha>temp?alpha:temp;
		if (alpha>=beta)
			return alpha; /* cut-off */
		b=alpha+1; /* set new null window */

		/* update the PV */
		pv[ply][ply]=move_list[i];
		int j;
		for (j = ply + 1; j < pv_length[ply + 1]; ++j)
			pv[ply][j]=pv[ply+1][j];
		pv_length[ply] = pv_length[ply + 1];
	}
	return alpha;
}

/* sort_pv() is called when the search function is following
   the PV (Principal Variation). It looks through the current
   ply's move list to see if the PV move is there. If so,
   it adds 10,000,000 to the move's score so it's played first
   by the search function. If not, follow_pv remains FALSE and
   search() stops calling sort_pv(). */

void sort_pv()
{
	int i;

	follow_pv = 0;
	for(i = first_move[ply]; i < first_move[ply + 1]; ++i)
		if
		(
			(move_list[i].from == pv[0][ply].from)
			&&(move_list[i].to == pv[0][ply].to)
			&&(move_list[i].piece_promote == pv[0][ply].piece_promote)
			&&(move_list[i].ep_location_created == pv[0][ply].ep_location_created)
			&&(move_list[i].ep_victim_location == pv[0][ply].ep_victim_location)
			&&(move_list[i].castle == pv[0][ply].castle)
		)
		{
			follow_pv = 1;
			move_list[i].score += 11000;
			return;
		}
}


/* sort() searches the current ply's move list from 'from'
   to the end to find the move with the highest score. Then it
   swaps that move and the 'from' move so the move with the
   highest score gets searched next, and hopefully produces
   a cutoff. */

void sort(int fromMove)
{
	int i;
	int bs;  /* best score */
	int bi;  /* best i */
	move* g=malloc(sizeof(move));

	bs = -1;
	bi = fromMove;
	for (i = fromMove; i < first_move[ply + 1]; ++i)
		if (move_list[i].score > bs) {
			bs = move_list[i].score;
			bi = i;
		}
	*g=move_list[fromMove];
	move_list[fromMove]=move_list[bi];
	move_list[bi]=*g;
	free(g);
}

void terminate()
{
	if (getTime()>=stop_time) {done=1;longjmp(env,0);}
}

int ProbeHash(int depth, int alpha, int beta)
{

    HASHE * phashe = &hash_table[abs(hash % HASH_SIZE)];

    if (phashe->key == hash) {
        if (phashe->depth >= depth) {
            if (phashe->flags == hashfEXACT)
                return phashe->value;
            if ((phashe->flags == hashfALPHA) &&
                (phashe->value <= alpha))
                return alpha;
            if ((phashe->flags == hashfBETA) &&
                (phashe->value >= beta))
                return beta;
        }
		//RememberBestMove();
    }
    return valUNKNOWN;
}

void RecordHash(int depth, int value, int flags)
{

    HASHE * phashe = &hash_table[abs(hash % HASH_SIZE)];

    phashe->key = hash;
    //phashe->best = BestMove();
    phashe->value = value;
    phashe->flags = flags;
    phashe->depth = depth;
}
